/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <string>
#include <vector>
#include <complex>
#include <math.h>
#include <algorithm>
//#include <unordered_map>

#include <net_types.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>
#include <BufferedNode.h>

////#include <boost/python.hpp>
////#include <boost/python/numeric.hpp>
////#include <boost/python/suite/indexing/vector_indexing_suite.hpp>
////#include <boost/python/suite/indexing/map_indexing_suite.hpp>
//#include <pybind11/eval.h>
//#include <pybind11/pybind11.h>
#include <pybind11/embed.h>
#include <pybind11/stl.h>
#include <pybind11/operators.h>
#include <pybind11/numpy.h>
//#include <pybind11/eigen.h>

namespace py = pybind11;
using namespace FD;

class PyCodeExecutor3;

DECLARE_NODE(PyCodeExecutor3);
/*Node
 *
 * @name PyCodeExecutor3
 * @category HARK:Python
 * @description This node executs an external python code. 
 *
 * @parameter_name DIRECTORY_NAME
 * @parameter_type string
 * @parameter_value .
 * @parameter_description [optional] The directory name of your python code. It will inserted to sys.path.
 *
 * @parameter_name MODULENAME
 * @parameter_type string
 * @parameter_value samplecode
 * @parameter_description Your python module name to import, i.e., your python code file name WITHOUT extension.
 * 
 * @parameter_name CLASSNAME
 * @parameter_type string
 * @parameter_value HarkNode
 * @parameter_description Your class name to call in this node.
 * 
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Debug option. If true, it prints input list and output list.
 * 
 END*/

class PyCodeExecutor3 : public BufferedNode {
  std::vector<int> inputsID;
  std::vector<std::string> inputsName; 
  std::vector<std::function<py::object(ObjectRef&)> > inputsMethod;

  std::vector<int> outputsID;
  std::vector<std::string> outputsName;
  std::vector<std::string> outputsType;
  std::vector<std::function<ObjectRef(py::dict&)> > outputsMethod;

  py::object main_namespace;
  std::string myNodeName;
  std::string classname;
  std::string filename;

  bool isDebug;
  bool isCalculated;
  std::string uuid;

  void uuidreplace(std::string& str, const std::string& from, const std::string& to){
    std::string::size_type pos = 0;
    while(pos = str.find(from, pos), pos != std::string::npos){
      str.replace(pos, from.length(), to);
      pos += to.length();
    }
  }

  inline bool starts_with(const std::string& s, const std::string& prefix){
    if(s.size() < prefix.size()){ return false; }
    return std::equal(prefix.begin(), prefix.end(), s.begin());
  }

  inline bool ends_with(const std::string& s, const std::string& suffix){
    if(s.size() < suffix.size()){ return false; }
    return std::equal(s.begin() + s.size() - suffix.size(), s.end(), suffix.begin());
  }

////
//  ////template<typename T>
//  ////void givePython(RCPtr<T> input, std::string inputName){
//    ////main_namespace[myNodeName.c_str()].attr(inputName.c_str()) = *input;
//  ////}
//
//  template<typename T>
//  py::array_t<T> getNumpyVector(ObjectRef& input) {
//    std::vector<T> o = object_cast<std::vector<T> >(input);
//    return py::array_t<T>({o.size()}, {sizeof(T)}, o.data());
//  }
//
//  template<typename T>
//  py::array_t<T> getNumpyMatrix(ObjectRef& input) {
//    Matrix<T> o = object_cast<Matrix<T> >(input);
//    return py::array_t<T>({o.nrows(), o.ncols()}, {sizeof(T)*o.ncols(), sizeof(T)}, o[0]);
//  }
////

public:
  PyCodeExecutor3(std::string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    // get parameters
    std::string dirname = "";
    if(parameters.exist("DIRECTORY_NAME")){
      dirname = object_cast<std::string>(parameters.get("DIRECTORY_NAME"));
    }
    filename = object_cast<std::string>(parameters.get("MODULENAME"));
    classname = object_cast<std::string>(parameters.get("CLASSNAME"));
    isDebug = dereference_cast<bool>(parameters.get("DEBUG"));
    isCalculated = false;

    //Py_Initialize();
    if(!Py_IsInitialized()){
      py::initialize_interpreter();
    }
    try{
      main_namespace = py::dict(py::module::import("__main__").attr("__dict__"));

      // get nodeID and name
      main_namespace["uuid"] = py::module::import("uuid");
      uuid = py::eval("str(uuid.uuid4())", main_namespace).cast<std::string>();
      uuidreplace(uuid, "-", "_");
      if(isDebug){ std::cerr << "[log] UUID:" << uuid << std::endl; }

      myNodeName = nodeName + uuid;
      int nodeID = -1;
      std::stringstream ss(nodeName.substr(nodeName.rfind("_")+1));
      ss >> nodeID;
      if(isDebug){ std::cerr <<  "[log] importing modules and numpy." << std::endl; }
      main_namespace["numpy"] = py::module::import("numpy");
      main_namespace["pprint"] = py::module::import("pprint");
      main_namespace["harkbasenode"] = py::module::import("harkpython.harkbasenode");
      
      // add the path for user's script
      main_namespace["sys"] = py::module::import("sys");
      if(dirname.length() > 0){
        py::list path = main_namespace["sys"].attr("path");
        path.append(py::cast(dirname.c_str()));
      }
      py::exec("print(sys.version_info)", main_namespace); // FOR DEBUG
      if(isDebug){
        py::exec("print(sys.path)\npprint.pprint(sys.path)", main_namespace);
      }

      if(isDebug){ std::cerr <<  "[log] importing harkpython." << std::endl; }
      py::exec("from harkpython import *", main_namespace);
      if(isDebug){ std::cerr <<  "[log] importing " << filename << "." << std::endl; }
      main_namespace[filename.c_str()] = py::module::import(filename.c_str());

      // instantiate user's code.
      main_namespace[myNodeName.c_str()] 
      	= main_namespace[filename.c_str()].attr(classname.c_str())();
      main_namespace[myNodeName.c_str()].attr("nodeID") = nodeID;
      main_namespace[myNodeName.c_str()].attr("nodeName") = nodeName;
      main_namespace[myNodeName.c_str()].attr("makeup")();
    }
    catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      throw (new NodeException(this, "Error in constructor", __FILE__, __LINE__));
    }
    inOrder = true;
  }

  ~PyCodeExecutor3(){
    try{
      main_namespace[myNodeName.c_str()].attr("__del__")();
    }
    catch(const py::error_already_set& e){
      if(isCalculated){
        std::cerr << e.what() << std::endl;
      	throw(new NodeException(this, "Error in destructor", __FILE__, __LINE__));      
      } else{}
    }
    //py::finalize_interpreter();
  }
  
  virtual int translateInput(const std::string &inputName){
    std::vector<std::string>::iterator result
      = find(inputsName.begin(), inputsName.end(), inputName);

    if(result != inputsName.end()){
      if(isDebug){
        std::cout << "translateInput already exists: (" << myNodeName << "::" << classname << ")\n"
                  << "Name: " << inputName << "(ID: " << result - inputsName.begin() << ")" << std::endl;
      }
      return inputsID[result - inputsName.begin()];
    }
    int id = addInput(inputName);
    inputsName.push_back(inputName);
    inputsID.push_back(id);

    if(isDebug){
      std::cout << "new translateInput: (" << myNodeName << "::" << classname << ")\n"
                << "Name: " << inputName << " (ID: " << id << ")" << std::endl;
    }

    return id;
  }
  
  virtual int translateOutput(const std::string &outputName){
    std::vector<std::string>::iterator result 
      = find(outputsName.begin(), outputsName.end(), outputName);
    
    if(result != outputsName.end()){
      if(isDebug){
        std::cout << "translateOutput already exists: (" << myNodeName << "::" << classname << ")\n" 
                  << "Name: " << outputName << " (ID: " << result - outputsName.begin() << ")" << std::endl;
      }
      return outputsID[result - outputsName.begin()];
    }
    int id = addOutput(outputName);
    outputsName.push_back(outputName);
    outputsID.push_back(id);

    if(isDebug){
      std::cout << "new translateOutput: (" << myNodeName << "::" << classname << ")\n" 
                << "Name: " << outputName << " (ID: " << id << ")" << std::endl;
    }

    py::tuple nam;
    py::tuple typ;
    try{
      nam = main_namespace[myNodeName.c_str()].attr("outputNames");
      typ = main_namespace[myNodeName.c_str()].attr("outputTypes");
    }
    catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      std::stringstream ss;
      ss << "Error during translate Output (" << filename << "::" << myNodeName << "::" << classname << ")"
         << "outputNames or outputTypes is not defined as tuple" << std::endl;
      throw (new NodeException(this, ss.str(),  __FILE__, __LINE__));
    }
    int inam = -1;
    try{
      inam = nam.attr("index")(outputName).cast<int>();
    }
    catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      std::stringstream ss;
      ss << "Error during translate Output (" << filename << "::" << myNodeName << "::" << classname << ")"
         << " Output Name: " << outputName << " of ID: " << id << " is not found in outputNames." << std::endl
         << "candidates are:::" << std::endl;
      for(size_t i=0; i<nam.size(); i++){
        std::string str = nam[i].cast<std::string>();
        ss << str << std::endl;
      }
      throw (new NodeException(this, ss.str(),  __FILE__, __LINE__));
    }
    std::string outputType = typ[inam].cast<std::string>();
    outputsType.push_back(outputType);

    // Registration of conversion lambda functions for output
    if(starts_with(outputType, "prim_")){
      if(outputType == "prim_null"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              return RCPtr<NilObject>(new NilObject());
            }
        );
      }
      else if(outputType == "prim_bool"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              return RCPtr<Bool>(new Bool(d[outputName.c_str()].cast<bool>()));
            }
        );
      }
      else if(outputType == "prim_int"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              return RCPtr<Int>(new Int(d[outputName.c_str()].cast<int>()));
            }
        );
      }
      else if(outputType == "prim_float"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              return RCPtr<Float>(new Float(d[outputName.c_str()].cast<float>()));
            }
        );
      }
      else if(outputType == "prim_complex"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              return RCPtr<Complex<float> >(new Complex<float>(d[outputName.c_str()].cast<std::complex<float> >()));
            }
        );
      }
      else if(outputType == "prim_source"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              RCPtr<Source> src = RCPtr<Source>(new Source);
              py::dict srcd = d[outputName.c_str()];
              src->id = srcd["id"].cast<int>();
              src->x[0] = ((py::list)srcd["x"])[0].cast<float>();
              src->x[1] = ((py::list)srcd["x"])[1].cast<float>();
              src->x[2] = ((py::list)srcd["x"])[2].cast<float>();
              src->power = srcd["power"].cast<float>();
              return src;
            }
        );
      }
      else if(outputType == "prim_string"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              return RCPtr<String>(new String(d[outputName.c_str()].cast<std::string>()));
            }
        );
      }
      else{
        std::stringstream ss;
        ss << "Unsupported type for output:" << outputType;
        throw new NodeException(this, ss.str(), __FILE__, __LINE__);
      }
    } // start_with(outputType, "prim_")

    else if(starts_with(outputType, "vector_")){
      if(outputType == "vector_bool"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmplist = d[outputName.c_str()];
              // Always generate Bool as Object.
              //RCPtr<Vector<bool> > frompy(new Vector<bool>());
              RCPtr<Vector<ObjectRef> > frompy(new Vector<ObjectRef>());
              for(size_t i=0; i<tmplist.size(); i++){
                //(*frompy).push_back(tmplist[i].cast<bool>());
                RCPtr<Bool> val = RCPtr<Bool>(new Bool(tmplist[i].cast<bool>()));
                (*frompy).push_back(val);
              }
              return frompy;
            }
        );
      }
      else if(outputType == "vector_int"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmplist = d[outputName.c_str()];
              RCPtr<Vector<int> > frompy(new Vector<int>());
              for(size_t i=0; i<tmplist.size(); i++){
                (*frompy).push_back(tmplist[i].cast<int>());
              }
              return frompy;
            }
        );
      }
      else if(outputType == "vector_float"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmplist = d[outputName.c_str()];
              RCPtr<Vector<float> > frompy(new Vector<float>());
              for(size_t i=0; i<tmplist.size(); i++){
                (*frompy).push_back(tmplist[i].cast<float>());
              }
              return frompy;
            }
        );
      }
      else if(outputType == "vector_complex"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmplist = d[outputName.c_str()];
              RCPtr<Vector<std::complex<float> > > frompy(new Vector<std::complex<float> >());
              for(size_t i=0; i<tmplist.size(); i++){
                (*frompy).push_back(tmplist[i].cast<std::complex<float> >());
              }
              return frompy;
            }
        );
      }
      else if(outputType == "vector_source"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmpvec = d[outputName.c_str()];
              RCPtr<Vector<ObjectRef> > frompy (new Vector<ObjectRef>());
              for(size_t i=0; i<tmpvec.size(); i++){
                RCPtr<Source> src = RCPtr<Source>(new Source);
                py::dict srcd = tmpvec[i];
                src->id = srcd["id"].cast<int>();
                src->x[0] = ((py::list)srcd["x"])[0].cast<float>();
                src->x[1] = ((py::list)srcd["x"])[1].cast<float>();
                src->x[2] = ((py::list)srcd["x"])[2].cast<float>();
                src->power = srcd["power"].cast<float>();
                (*frompy).push_back(src);
              }
              return frompy;
            }
        );
      }
      else{
        std::stringstream ss;
        ss << "Unsupported type for output:" << outputType;
        throw new NodeException(this, ss.str(), __FILE__, __LINE__);
      }
    } // start_with(outputType, "vector_")

    else if(starts_with(outputType, "matrix_")){
      if(outputType == "matrix_bool"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmpmat = d[outputName.c_str()];
              // Always generate Bool as Object.
              //RCPtr<Matrix<bool> > frompy(new Matrix<bool>(tmpmat.size(), ((py::list)tmpmat[0]).size()));
              RCPtr<Matrix<ObjectRef> > frompy(new Matrix<ObjectRef>(tmpmat.size(), ((py::list)tmpmat[0]).size()));
              for(size_t r=0; r<tmpmat.size(); r++){
                for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
                  //(*frompy)(r, c) = ((py::list)tmpmat[r])[c].cast<bool>();
                  RCPtr<Bool> val = RCPtr<Bool>(new Bool(((py::list)tmpmat[r])[c].cast<bool>()));
                  (*frompy)(r, c) = val;
                }
              }
              return frompy;
            }
        );
      }
      else if(outputType == "matrix_int"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmpmat = d[outputName.c_str()];
              RCPtr<Matrix<int> > frompy(new Matrix<int>(tmpmat.size(), ((py::list)tmpmat[0]).size()));
              for(size_t r=0; r<tmpmat.size(); r++){
                for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
                  (*frompy)(r, c) = ((py::list)tmpmat[r])[c].cast<int>();
                }
              }
              return frompy;
            }
        );
      }  
      else if(outputType == "matrix_float"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmpmat = d[outputName.c_str()];
              RCPtr<Matrix<float> > frompy(new Matrix<float>(tmpmat.size(), ((py::list)tmpmat[0]).size()));
              for(size_t r=0; r<tmpmat.size(); r++){
                for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
                  (*frompy)(r, c) = ((py::list)tmpmat[r])[c].cast<float>();
                }
              }
              return frompy;
            }
        );
      }
      else if(outputType == "matrix_complex"){
        outputsMethod.push_back(
            [outputName](py::dict &d){
              py::list tmpmat = d[outputName.c_str()];
              RCPtr<Matrix<std::complex<float> > > frompy(new Matrix<std::complex<float> >(tmpmat.size(), ((py::list)tmpmat[0]).size()));
              for(size_t r=0; r<tmpmat.size(); r++){
                for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
                  (*frompy)(r, c) = ((py::list)tmpmat[r])[c].cast<std::complex<float> >();
                }
              }
              return frompy;
            }
        );
      }
      else{
        std::stringstream ss;
        ss << "Unsupported type for output:" << outputType;
        throw new NodeException(this, ss.str(), __FILE__, __LINE__);
      }
    } // start_with(outputType, "matrix_")

    else if(starts_with(outputType, "map_")){
      if(starts_with(outputType, "map_vector_")){
        if(outputType == "map_vector_int"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  RCPtr<Vector<int> > value = RCPtr<Vector<int> >(new Vector<int>);
                  py::list tmplist = map[keys[i]];
                  for(size_t j=0; j<tmplist.size(); j++){
                    (*value).push_back(tmplist[j].cast<int>());
                  }
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else if(outputType == "map_vector_float"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  RCPtr<Vector<float> > value = RCPtr<Vector<float> >(new Vector<float>);
                  py::list tmplist = map[keys[i]];
                  for(size_t j=0; j<tmplist.size(); j++){
                    (*value).push_back(tmplist[j].cast<float>());
                  }
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else if(outputType == "map_vector_complex"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy(new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  RCPtr<Vector<std::complex<float> > > value = RCPtr<Vector<std::complex<float> > >(new Vector<std::complex<float> >);
                  py::list tmplist = map[keys[i]];
                  for(size_t j=0; j<tmplist.size(); j++){
                    (*value).push_back(tmplist[j].cast<std::complex<float> >());
                  }
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else{
          std::stringstream ss;
          ss << "Unsupported type for output:" << outputType;
          throw new NodeException(this, ss.str(), __FILE__, __LINE__);
        }
      } // start_with(outputType, "map_vector_")

      else if(starts_with(outputType, "map_matrix_")){
        if(outputType == "map_matrix_int"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  py::list tmpmat = map[keys[i]];
                  RCPtr<Matrix<int> > value = RCPtr<Matrix<int> >(new Matrix<int>(tmpmat.size(), ((py::list)tmpmat[0]).size()));
                  for(size_t r=0; r<tmpmat.size(); r++){
                    for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
                      (*value)(r, c) = ((py::list)tmpmat[r])[c].cast<int>();
                    }
                  }
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else if(outputType == "map_matrix_float"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  py::list tmpmat = map[keys[i]];
                  RCPtr<Matrix<float> > value = RCPtr<Matrix<float> >(new Matrix<float>(tmpmat.size(), ((py::list)tmpmat[0]).size()));
                  for(size_t r=0; r<tmpmat.size(); r++){
                    for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
                      (*value)(r, c) = ((py::list)tmpmat[r])[c].cast<float>();
                    }
                  }
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else if(outputType == "map_matrix_complex"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy(new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  py::list tmpmat = map[keys[i]];
                  RCPtr<Matrix<std::complex<float> > > value = RCPtr<Matrix<std::complex<float> > >(new Matrix<std::complex<float> >(tmpmat.size(), ((py::list)tmpmat[0]).size()));
                  for(size_t r=0; r<tmpmat.size(); r++){
                    for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
                      (*value)(r, c) = ((py::list)tmpmat[r])[c].cast<std::complex<float> >();
                    }
                  }
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else{
          std::stringstream ss;
          ss << "Unsupported type for output:" << outputType;
          throw new NodeException(this, ss.str(), __FILE__, __LINE__);
        }
      } // start_with(outputType, "map_matrix_")

      else if(starts_with(outputType, "map_prime_")){
        if(outputType == "map_prim_bool"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  RCPtr<Bool> value = RCPtr<Bool>(new Bool(map[keys[i]].cast<bool>()));
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else if(outputType == "map_prim_int"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  RCPtr<Int> value = RCPtr<Int>(new Int(map[keys[i]].cast<int>()));
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else if(outputType == "map_prim_float"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  RCPtr<Float> value = RCPtr<Float>(new Float(map[keys[i]].cast<float>()));
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else if(outputType == "map_prim_complex"){
          outputsMethod.push_back(
              [outputName](py::dict &d){
                py::dict map = d[outputName.c_str()];
                py::list keys = map.attr("keys")();
                RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
                for(size_t i=0; i<keys.size(); i++){
                  RCPtr<Complex<float> > value = RCPtr<Complex<float> >(new Complex<float>(map[keys[i]].cast<std::complex<float> >()));
                  (*frompy)[keys[i].cast<int>()] = value;
                }
                return frompy;
              }
          );
        }
        else{
          std::stringstream ss;
          ss << "Unsupported type for output:" << outputType;
          throw new NodeException(this, ss.str(), __FILE__, __LINE__);
        }
      } // start_with(outputType, "map_prim_")

/*
      else if(starts_with(outputType, "map_string_")){
        //else if(outputType.find("map_string_map_string:", 0) != std::string::npos){
        if(outputType.find("map_string_map_string:", 0) != std::string::npos){
          unsigned long long int index, offset;
          std::map<unsigned long long int, std::string> typesMap;
          index = outputType.find("map_string_map_string:", 0) + 22;
          offset = index;
          while(offset != std::string::npos){
            offset = outputType.find("prim_int", offset);
            if(offset != std::string::npos){
              typesMap[offset]="prim_int";
              offset+=8;
            }
            //std::cerr << "prim_int: count" << count << " offset" << offset << " npos" << std::string::npos << std::endl;
          }
          offset = index;
          while(offset != std::string::npos){
            offset = outputType.find("prim_float", offset);
            if(offset != std::string::npos){
              typesMap[offset]="prim_float";
              offset+=10;
            }
            //std::cerr << "prim_float: count" << count << " offset" << offset << " npos" << std::string::npos << std::endl;
          }
          std::map<unsigned long long int, std::string>::iterator it = typesMap.begin();
          py::dict map = d[(*outputsNameit).c_str()];
          py::list keys = map.attr("keys")();
          keys.attr("sort")();
          RCPtr<Map<std::string, Map<std::string, ObjectRef> > > frompy = RCPtr<Map<std::string, Map<std::string, ObjectRef> > >(new Map<std::string, Map<std::string, ObjectRef> >);
          for(size_t i=0; i<keys.size(); i++){
            RCPtr<Map<std::string, ObjectRef> > mapinmap = RCPtr<Map<std::string, ObjectRef> >(new Map<std::string, ObjectRef>);
            py::dict mapmap = map[keys[i]];
            py::list mapkeys = mapmap.attr("keys")();
            mapkeys.attr("sort")();
            for(size_t j=0; j<mapkeys.size(); j++){
              if(it->second == "prim_int"){
                RCPtr<Int> value = RCPtr<Int>(new Int);
                (*mapinmap)[mapkeys[j].cast<std::string>()] = value;
              }
              else if(it->second == "prim_float"){
                RCPtr<Float> value = RCPtr<Float>(new Float);
                (*mapinmap)[mapkeys[j].cast<std::string>()] = value;
              }
              it++;
            }
            (*frompy)[keys[i].cast<std::string>()] = *mapinmap;
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
      } // start_with(outputType, "map_string_")
*/
    } // start_with(outputType, "map_")

    else{
      std::stringstream ss;
      ss << "Unsupported type for output:" << outputType;
      throw new NodeException(this, ss.str(), __FILE__, __LINE__);
    }

    if(isDebug){
      std::cout << "got outputsMethod: (" << myNodeName << "::" << classname << ")\n" 
                << "Name: " << outputName << " (ID: " << id << ")" << std::endl;
    }
////////////////////////////////////////

    return id;
  }
  
  virtual void initialize(void)
  {
    int maxahead = 0, maxback = 0;
    std::vector<int>::iterator it = outputsID.begin();
    while(it != outputsID.end()){
      if(maxahead < outputs[*it].lookAhead){ maxahead = outputs[*it].lookAhead; }
      if(maxback < outputs[*it].lookBack){ maxback = outputs[*it].lookBack; }
      it++;
    }
    maxahead++;
    maxback++;
    it = outputsID.begin();
    while(it != outputsID.end()){
      outputs[*it].lookAhead = maxahead;
      outputs[*it].lookBack = maxback;
      it++;
    }
    this->BufferedNode::initialize();
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    // get input
    if(count == 0){
      std::vector<std::string>::iterator inputsNameit = inputsName.begin();
      std::vector<int>::iterator inputsIDit = inputsID.begin();

      // Registration of conversion lambda functions for input
      while(inputsIDit != inputsID.end()){
        ObjectRef input = getInput(*inputsIDit, count);

        // give data to python depending on their types 
        //// Primitives of ObjectRef
        if(typeid(*input) == typeid(NilObject)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::none();
              }
          );
        }
        else if(typeid(*input) == typeid(Bool)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(dereference_cast<bool>(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(Int)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(dereference_cast<int>(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(Float)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(dereference_cast<float>(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(Double)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(dereference_cast<double>(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(Complex<float>)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(object_cast<std::complex<float> >(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(Complex<double>)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(object_cast<std::complex<double> >(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(Source)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                Source src = object_cast<Source>(obj);
                py::dict pySource = py::dict();
                pySource["id"] = src.id;
                std::array<float, 3> arrX = {src.x[0], src.x[1], src.x[2]};
                pySource["x"] = arrX;
                pySource["power"] = src.power;
                return pySource;
              }
          );
        }
        else if(typeid(*input) == typeid(Char)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(dereference_cast<char>(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(String)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(object_cast<std::string>(obj));
              }
          );
        }

        //// Vector<> container of ObjectRef
        else if(typeid(*input) == typeid(Vector<bool>)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(object_cast<std::vector<bool> >(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(Vector<int>)){
          inputsMethod.push_back(
/*
              [this](ObjectRef &obj) -> py::object {
                return getNumpyVector<int>(obj);
              }
*/
              [](ObjectRef &obj) -> py::object {
                std::vector<int> o = object_cast<std::vector<int> >(obj);
                return py::array_t<int>({o.size()}, {sizeof(int)}, o.data());
              }
          );
        }
        else if(typeid(*input) == typeid(Vector<float>)){
          inputsMethod.push_back(
/*
              [this](ObjectRef &obj) -> py::object {
                return getNumpyVector<float>(obj);
              }
*/
              [](ObjectRef &obj) -> py::object {
                std::vector<float> o = object_cast<std::vector<float> >(obj);
                return py::array_t<float>({o.size()}, {sizeof(float)}, o.data());
              }
          );
        }
        else if(typeid(*input) == typeid(Vector<double>)){
          inputsMethod.push_back(
/*
              [this](ObjectRef &obj) -> py::object {
                return getNumpyVector<double>(obj);
              }
*/
              [](ObjectRef &obj) -> py::object {
                std::vector<double> o = object_cast<std::vector<double> >(obj);
                return py::array_t<double>({o.size()}, {sizeof(double)}, o.data());
              }
          );
        }
        else if(typeid(*input) == typeid(Vector<std::complex<float> >)){
          inputsMethod.push_back(
/*
              [this](ObjectRef &obj) -> py::object {
                return getNumpyVector<std::complex<float> >(obj);
              }
*/
              [](ObjectRef &obj) -> py::object {
                std::vector<std::complex<float> > o = object_cast<std::vector<std::complex<float> > >(obj);
                return py::array_t<std::complex<float> >({o.size()}, {sizeof(std::complex<float>)}, o.data());
              }
          );
        }
        else if(typeid(*input) == typeid(Vector<char>)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                return py::cast(object_cast<std::vector<char> >(obj));
              }
          );
        }
        else if(typeid(*input) == typeid(Vector<ObjectRef>)){
          Vector<ObjectRef> tmp = object_cast<Vector<ObjectRef> >(input);

          if(tmp.size()==0){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  return py::list();
                }
            );
          }
          else{
            if(typeid(*(tmp[0])) == typeid(Bool)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> >(obj);
                    py::list topylist = py::list(vec.size());
                    for(unsigned int i = 0; i < vec.size(); i++){
                      topylist[i] = dereference_cast<bool>(vec[i]);
                    }
                    return topylist;
                  }
              );
            }
            else if(typeid(*(tmp[0])) == typeid(Int)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> >(obj);
                    std::vector<int> v(vec.size());
                    for(int i = 0; i < vec.size(); i++){
                      v[i] = dereference_cast<int>(vec[i]);
                    }
                    return py::array_t<int>({v.size()}, {sizeof(int)}, v.data());
                  }
              );
            }
            else if(typeid(*(tmp[0])) == typeid(Float)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> >(obj);
                    std::vector<float> v(vec.size());
                    for(int i = 0; i < vec.size(); i++){
                      v[i] = dereference_cast<float>(vec[i]);
                    }
                    return py::array_t<float>({v.size()}, {sizeof(float)}, v.data());
                  }
              );
            }
            else if(typeid(*(tmp[0])) == typeid(Double)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> >(obj);
                    std::vector<double> v(vec.size());
                    for(int i = 0; i < vec.size(); i++){
                      v[i] = dereference_cast<double>(vec[i]);
                    }
                    return py::array_t<double>({v.size()}, {sizeof(double)}, v.data());
                  }
              );
            }
            else if(typeid(*(tmp[0])) == typeid(Complex<float>)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> >(obj);
                    std::vector<std::complex<float> > v(vec.size());
                    for(int i = 0; i < vec.size(); i++){
                      v[i] = object_cast<std::complex<float> >(vec[i]);
                    }
                    return py::array_t<std::complex<float> >({v.size()}, {sizeof(std::complex<float>)}, v.data());
                  }
              );
            }
            else if(typeid(*(tmp[0])) == typeid(Char)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> >(obj);
                    py::list topylist = py::list(vec.size());
                    for(unsigned int i = 0; i < vec.size(); i++){
                      topylist[i] = dereference_cast<char>(vec[i]);
                    }
                    return topylist;
                  }
              );
            }
            else if(typeid(*(tmp[0])) == typeid(Source)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> >(obj);
                    py::list topylist = py::list(vec.size());
                    for(unsigned int i = 0; i < vec.size(); i++){
                      Source src = object_cast<Source>(vec[i]);
                      py::dict pySource = py::dict();
                      pySource["id"] = src.id;
                      std::array<float, 3> arrX = {src.x[0], src.x[1], src.x[2]};
                      pySource["x"] = arrX;
                      pySource["power"] = src.power;
                      topylist[i] = pySource;
                    }
                    return topylist;
                  }
              );
            }
            else{
              std::stringstream ss;
              ss << "Input of Vector<ObjectRef> [ObjectRef = " << typeid(*(tmp[0])).name() << "] is not supported";
              throw new NodeException(this, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
        }

        //// Matrix<> container of ObjectRef
        else if(typeid(*input) == typeid(Matrix<bool>)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                Matrix<bool> input_mat = object_cast<Matrix<bool> >(obj);
                std::vector<std::vector<bool> > mat(input_mat.nrows(), std::vector<bool>(input_mat.ncols()));
                for(int r = 0; r < input_mat.nrows(); r++){
                  for(int c = 0; c < input_mat.ncols(); c++){
                    mat[r][c] = input_mat(r, c);
                  }
                }
                return py::cast(mat);
              }
          );
        }
        else if(typeid(*input) == typeid(Matrix<int>)){
          inputsMethod.push_back(
/*
              [this](ObjectRef &obj) -> py::object {
                return getNumpyMatrix<int>(obj);
              }
*/
              [](ObjectRef &obj) -> py::object {
                Matrix<int> o = object_cast<Matrix<int> >(obj);
                return py::array_t<int>({o.nrows(), o.ncols()}, {sizeof(int)*o.ncols(), sizeof(int)}, o[0]);
              }
          );
        }
        else if(typeid(*input) == typeid(Matrix<float>)){
          inputsMethod.push_back(
/*
              [this](ObjectRef &obj) -> py::object {
                return getNumpyMatrix<float>(obj);
              }
*/
              [](ObjectRef &obj) -> py::object {
                Matrix<float> o = object_cast<Matrix<float> >(obj);
                return py::array_t<float>({o.nrows(), o.ncols()}, {sizeof(float)*o.ncols(), sizeof(float)}, o[0]);
              }
          );
        }
        else if(typeid(*input) == typeid(Matrix<double>)){
          inputsMethod.push_back(
/*
              [this](ObjectRef &obj) -> py::object {
                return getNumpyMatrix<double>(obj);
              }
*/
              [](ObjectRef &obj) -> py::object {
                Matrix<double> o = object_cast<Matrix<double> >(obj);
                return py::array_t<double>({o.nrows(), o.ncols()}, {sizeof(double)*o.ncols(), sizeof(double)}, o[0]);
              }
          );
        }
        else if(typeid(*input) == typeid(Matrix<std::complex<float> >)){
          inputsMethod.push_back(
/*
              [this](ObjectRef &obj) -> py::object {
                return getNumpyMatrix<std::complex<float> >(obj);
              }
*/
              [](ObjectRef &obj) -> py::object {
                Matrix<std::complex<float> > o = object_cast<Matrix<std::complex<float> > >(obj);
                return py::array_t<std::complex<float> >({o.nrows(), o.ncols()}, {sizeof(std::complex<float>)*o.ncols(), sizeof(std::complex<float>)}, o[0]);
              }
          );
        }
        else if(typeid(*input) == typeid(Matrix<char>)){
          inputsMethod.push_back(
              [](ObjectRef &obj) -> py::object {
                Matrix<char> input_mat = object_cast<Matrix<char> >(obj);
                std::vector<std::vector<char> > mat(input_mat.nrows(), std::vector<char>(input_mat.ncols()));
                for(int r = 0; r < input_mat.nrows(); r++){
                  for(int c = 0; c < input_mat.ncols(); c++){
                    mat[r][c] = input_mat(r, c);
                  }
                }
                return py::cast(mat);
              }
          );
        }
        else if(typeid(*input) == typeid(Matrix<ObjectRef>)){
          Matrix<ObjectRef> tmp = object_cast<Matrix<ObjectRef> >(input);

          if(tmp.nrows()==0 || tmp.ncols()==0){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  return py::list();
                }
            );
          }
          else{
            if(typeid(*(tmp(0,0))) == typeid(Bool)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Matrix<ObjectRef> mat = object_cast<Matrix<ObjectRef> >(obj);
                    py::list topylist = py::list(mat.nrows());
                    for(int r = 0; r < mat.nrows(); r++){
                      py::list vlist = py::list(mat.ncols());
                      for(int c = 0; c < mat.ncols(); c++){
                        vlist[c] = dereference_cast<bool>(mat(r, c));
                      }
                      topylist[r] = vlist;
                    }
                    return topylist;
                  }
              );
            }
            else if(typeid(*(tmp(0,0))) == typeid(Int)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Matrix<ObjectRef> mat = object_cast<Matrix<ObjectRef> >(obj);
                    Matrix<int> m(mat.nrows(), mat.ncols());
                    for(int r = 0; r < mat.nrows(); r++){
                      for(int c = 0; c < mat.ncols(); c++){
                        m(r, c) = dereference_cast<int>(mat(r, c));
                      }
                    }
                    return py::array_t<int>({m.nrows(), m.ncols()}, {sizeof(int)*m.ncols(), sizeof(int)}, m[0]);
                  }
              );
            }
            else if(typeid(*(tmp(0,0))) == typeid(Float)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Matrix<ObjectRef> mat = object_cast<Matrix<ObjectRef> >(obj);
                    Matrix<float> m(mat.nrows(), mat.ncols());
                    for(int r = 0; r < mat.nrows(); r++){
                      for(int c = 0; c < mat.ncols(); c++){
                        m(r, c) = dereference_cast<float>(mat(r, c));
                      }
                    }
                    return py::array_t<float>({m.nrows(), m.ncols()}, {sizeof(float)*m.ncols(), sizeof(float)}, m[0]);
                  }
              );
            }
            else if(typeid(*(tmp(0,0))) == typeid(Double)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Matrix<ObjectRef> mat = object_cast<Matrix<ObjectRef> >(obj);
                    Matrix<double> m(mat.nrows(), mat.ncols());
                    for(int r = 0; r < mat.nrows(); r++){
                      for(int c = 0; c < mat.ncols(); c++){
                        m(r, c) = dereference_cast<double>(mat(r, c));
                      }
                    }
                    return py::array_t<double>({m.nrows(), m.ncols()}, {sizeof(double)*m.ncols(), sizeof(double)}, m[0]);
                  }
              );
            }
            else if(typeid(*(tmp(0,0))) == typeid(Complex<float>)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Matrix<ObjectRef> mat = object_cast<Matrix<ObjectRef> >(obj);
                    Matrix<std::complex<float> > m(mat.nrows(), mat.ncols());
                    for(int r = 0; r < mat.nrows(); r++){
                      for(int c = 0; c < mat.ncols(); c++){
                        m(r, c) = object_cast<std::complex<float> >(mat(r, c));
                      }
                    }
                    return py::array_t<std::complex<float> >({m.nrows(), m.ncols()}, {sizeof(std::complex<float>)*m.ncols(), sizeof(std::complex<float>)}, m[0]);
                  }
              );
            }
            else if(typeid(*(tmp(0,0))) == typeid(Char)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Matrix<ObjectRef> mat = object_cast<Matrix<ObjectRef> >(obj);
                    py::list topylist = py::list(mat.nrows());
                    for(int r = 0; r < mat.nrows(); r++){
                      py::list vlist = py::list(mat.ncols());
                      for(int c = 0; c < mat.ncols(); c++){
                        vlist[c] = dereference_cast<char>(mat(r, c));
                      }
                      topylist[r] = vlist;
                    }
                    return topylist;
                  }
              );
            }
            else if(typeid(*(tmp(0,0))) == typeid(Source)){
              inputsMethod.push_back(
                  [](ObjectRef &obj) -> py::object {
                    Matrix<ObjectRef> mat = object_cast<Matrix<ObjectRef> >(obj);
                    py::list topylist = py::list(mat.nrows());
                    for(int r = 0; r < mat.nrows(); r++){
                      py::list vlist = py::list(mat.ncols());
                      for(int c = 0; c < mat.ncols(); c++){
                        Source src = object_cast<Source>(mat(r, c));
                        py::dict pySource = py::dict();
                        pySource["id"] = src.id;
                        std::array<float, 3> arrX = {src.x[0], src.x[1], src.x[2]};
                        pySource["x"] = arrX;
                        pySource["power"] = src.power;
                        vlist[c]=pySource;
                      }
                      topylist[r] = vlist;
                    }
                    return topylist;
                  }
              );
            }
            else{
              std::stringstream ss;
              ss << "Input of Matrix<ObjectRef> [ObjectRef = " << typeid(*(tmp(0, 0))).name() << "] is not supported";
              throw new NodeException(this, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
        }

        //// Map<>
        else if(typeid(*input) == typeid(Map<int, ObjectRef>)){
          Map<int, ObjectRef> tmp = object_cast<Map<int, ObjectRef> >(input);
          Map<int, ObjectRef >::iterator tmpit = tmp.begin();

          if(tmp.begin() == tmp.end()){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  return py::dict();
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Bool)){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, bool>();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, dereference_cast<bool>(it->second));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Int)){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, int>();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, dereference_cast<int>(it->second));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Float)){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, float>();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, dereference_cast<float>(it->second));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Double)){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, double>();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, dereference_cast<double>(it->second));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Complex<float>)){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, std::complex<float> >();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, dereference_cast<std::complex<float> >(it->second));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Complex<double>)){
            inputsMethod.push_back(
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, std::complex<double> >();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, dereference_cast<std::complex<double> >(it->second));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Vector<int>)){
            inputsMethod.push_back(
/*
                [this](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<int> >();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, getNumpyVector<int>(it->second));
                  }
                  return py::cast(topymap);
                }
*/
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<int> >();
                  for(; it != topy.end(); it++) {
                    std::vector<int> o = object_cast<std::vector<int> >(it->second);
                    topymap.emplace_hint(topymap.end(), it->first, py::array_t<int>({o.size()}, {sizeof(int)}, o.data()));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Vector<float>)){
            inputsMethod.push_back(
/*
                [this](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<float> >();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, getNumpyVector<float>(it->second));
                  }
                  return py::cast(topymap);
                }
*/
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<float> >();
                  for(; it != topy.end(); it++) {
                    std::vector<float> o = object_cast<std::vector<float> >(it->second);
                    topymap.emplace_hint(topymap.end(), it->first, py::array_t<float>({o.size()}, {sizeof(float)}, o.data()));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Vector<std::complex<float> >)){
            inputsMethod.push_back(
/*
                [this](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<std::complex<float> > >();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, getNumpyVector<std::complex<float> >(it->second));
                  }
                  return py::cast(topymap);
                }
*/
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<std::complex<float> > >();
                  for(; it != topy.end(); it++) {
                    std::vector<std::complex<float> > o = object_cast<std::vector<std::complex<float> > >(it->second);
                    topymap.emplace_hint(topymap.end(), it->first, py::array_t<std::complex<float> >({o.size()}, {sizeof(std::complex<float>)}, o.data()));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Matrix<int>)){
            inputsMethod.push_back(
/*
                [this](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<int> >();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, getNumpyMatrix<int>(it->second));
                  }
                  return py::cast(topymap);
                }
*/
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<int> >();
                  for(; it != topy.end(); it++) {
                    Matrix<int> o = object_cast<Matrix<int> >(it->second);
                    topymap.emplace_hint(topymap.end(), it->first, py::array_t<int>({o.nrows(), o.ncols()}, {sizeof(int)*o.ncols(), sizeof(int)}, o[0]));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Matrix<float>)){
            inputsMethod.push_back(
/*
                [this](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<float> >();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, getNumpyMatrix<float>(it->second));
                  }
                  return py::cast(topymap);
                }
*/
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<float> >();
                  for(; it != topy.end(); it++) {
                    Matrix<float> o = object_cast<Matrix<float> >(it->second);
                    topymap.emplace_hint(topymap.end(), it->first, py::array_t<float>({o.nrows(), o.ncols()}, {sizeof(float)*o.ncols(), sizeof(float)}, o[0]));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else if(typeid(*(tmpit->second)) == typeid(Matrix<std::complex<float> >)){
            inputsMethod.push_back(
/*
                [this](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<std::complex<float> > >();
                  for(; it != topy.end(); it++) {
                    topymap.emplace_hint(topymap.end(), it->first, getNumpyMatrix<std::complex<float> >(it->second));
                  }
                  return py::cast(topymap);
                }
*/
                [](ObjectRef &obj) -> py::object {
                  Map<int, ObjectRef> topy = object_cast<Map<int, ObjectRef> >(obj);
                  Map<int, ObjectRef >::iterator it = topy.begin();
                  auto topymap = std::map<int, py::array_t<std::complex<float> > >();
                  for(; it != topy.end(); it++) {
                    Matrix<std::complex<float> > o = object_cast<Matrix<std::complex<float> > >(it->second);
                    topymap.emplace_hint(topymap.end(), it->first, py::array_t<std::complex<float> >({o.nrows(), o.ncols()}, {sizeof(std::complex<float>)*o.ncols(), sizeof(std::complex<float>)}, o[0]));
                  }
                  return py::cast(topymap);
                }
            );
          }
          else{
            std::stringstream ss;
            ss << "Map value" << typeid(*(tmpit->second)).name() << " is not supported";
            throw new NodeException(this, ss.str().c_str(), __FILE__, __LINE__);
          }
        }
/*
        else if(typeid(*input) == typeid(Map<std::string, ObjectRef>)){
          Map<std::string, ObjectRef>  topy = object_cast<Map<std::string, ObjectRef> > (input);
          py::dict topydict = py::dict();

          Map<std::string, ObjectRef >::iterator it = topy.begin();
          for(; it != topy.end(); it++){
            if(typeid(*(it->second)) == typeid(Map<std::string, int>)){
              topydict[it->first.c_str()] = object_cast<Int>(it->second);
            }
            else if(typeid(*(it->second)) == typeid(Map<std::string, float>)){
              topydict[it->first.c_str()] = object_cast<Float>(it->second);
            }
            else if(typeid(*(it->second)) == typeid(Map<std::string, std::string>)){
              topydict[it->first.c_str()] = object_cast<String>(it->second);
            }
            else{
              std::stringstream ss;
              ss << "Map<std::string, T> value" << typeid(*(it->second)).name() << " is not supported";
              throw new NodeException(this, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
          return topydict; 
        }
*/
        else{
          if(isDebug){ std::cout << std::endl; }
          std::stringstream ss;
          ss << "Type (" << typeid(*input).name() << ") of input terminal "
             << *inputsNameit << " is unsupported." << std::endl;
          throw new NodeException (this, ss.str(), __FILE__, __LINE__); 
        }

        if(isDebug){
          std::cout << "got inputsMethod (" << myNodeName << "::" << classname << ")\n"
                    << "Name: " << *inputsNameit << " (ID: " << *inputsIDit << ")" << std::endl;
        }

        inputsIDit++;
        inputsNameit++;
      }
    } // if(count == 0)

    {
      std::vector<std::string>::iterator inputsNameit = inputsName.begin();
      std::vector<int>::iterator inputsIDit = inputsID.begin();
      std::vector<std::function<py::object(ObjectRef&)> >::iterator inputsMethodit = inputsMethod.begin();

////
//    try{
////
      while(inputsIDit != inputsID.end()){
        ObjectRef input = getInput(*inputsIDit, count);
        if(isDebug){
          std::cout << "Input terminal: (" << myNodeName << "::" << classname << ")\n"
                    << "Name: " << *inputsNameit << " (ID: " << *inputsIDit << ")" << std::endl;
        }
        main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = (*inputsMethodit)(input);

        inputsIDit++;
        inputsNameit++;
        inputsMethodit++;
      }
////
//    }
//    catch(const py::error_already_set& e){
//      std::cerr << e.what() << std::endl;
//      throw (new NodeException(this, "Error during getting inputs", __FILE__, __LINE__));
//    }
////
    }
    
    //calculate
    {
      if(isDebug){
        std::cerr << "calculating..." << std::endl;
      }

////
//    try{
////
      main_namespace[myNodeName.c_str()].attr("count") = count;
      py::exec((myNodeName + ".calculate()"), main_namespace);
////
//    }
//    catch(const py::error_already_set& e){
//      std::cerr << e.what() << std::endl;
//      throw(new NodeException(this, "Error during calculate function execution", __FILE__, __LINE__));
//    }
////
    }    

    //set output
    {
      std::vector<std::string>::iterator outputsNameit = outputsName.begin();
      std::vector<int>::iterator outputsIDit = outputsID.begin();
      std::vector<std::string>::iterator outputsTypeit = outputsType.begin();
      std::vector<std::function<ObjectRef(py::dict&)> >::iterator outputsMethodit = outputsMethod.begin();

////
//    try{
////
      py::dict d = main_namespace[myNodeName.c_str()].attr("outputValues");
      
      while(outputsIDit != outputsID.end()){
        if(isDebug){
          std::cout << "Output terminal (" << myNodeName << "::" << classname << ")  " << "Name: " << *outputsNameit
                    << ", Type: " << *outputsTypeit << ", ID: " << *outputsIDit << std::endl;
        }
        (*(outputs[*outputsIDit].buffer))[count] = (*outputsMethodit)(d);

        outputsNameit++;
        outputsIDit++;
        outputsTypeit++;
        outputsMethodit++;
      }
////
//    }
//    catch(const py::error_already_set& e){
//      std::cerr << e.what() << std::endl; 
//      std::stringstream ss;
//      ss << "Error during setting outputs of type (" << *outputsTypeit << ") " << "and name (" << *outputsNameit << ")"
//         << std::endl;
//      throw (new NodeException(this, ss.str(), __FILE__, __LINE__));
//    }
//    catch(const std::exception& e){
//      std::cerr << e.what() << std::endl;
//      std::stringstream ss;
//      ss << "Error during setting outputs of type (" << *outputsTypeit << ") " << "and name (" << *outputsNameit << ")"
//         << std::endl;
//      throw (new NodeException(this, ss.str(), __FILE__, __LINE__));
//    }
////
    }

    isCalculated = true;
  }

  IN_ORDER_NODE_SPEEDUP(PyCodeExecutor3);
};
